from autolens import decorator_util
import numpy as np

from autolens.array.util import mask_util
from autolens.array.mapping_util import array_mapping_util, grid_mapping_util


@decorator_util.jit()
def centres_from_shape_pixel_scales_and_origin(shape, pixel_scales, origin):
    """Determine the (y,x) arc-second central coordinates of an array from its shape, pixel-scales and origin.

     The coordinate system is defined such that the positive y axis is up and positive x axis is right.

    Parameters
     ----------
    shape : (int, int)
        The (y,x) shape of the 2D array the arc-second centre is computed for.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the 2D array.
    origin : (float, flloat)
        The (y,x) origin of the 2D array, which the centre is shifted to.

    Returns
    --------
    tuple (float, float)
        The (y,x) arc-second central coordinates of the input array.

    Examples
    --------
    centres_arcsec = centres_from_shape_pixel_scales_and_origin(shape=(5,5), pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    y_centre_arcsec = float(shape[0] - 1) / 2 + (origin[0] / pixel_scales[0])
    x_centre_arcsec = float(shape[1] - 1) / 2 - (origin[1] / pixel_scales[1])

    return (y_centre_arcsec, x_centre_arcsec)


@decorator_util.jit()
def grid_centre_from_grid_1d(grid_1d):
    centre_y = (np.max(grid_1d[:, 0]) + np.min(grid_1d[:, 0])) / 2.0
    centre_x = (np.max(grid_1d[:, 1]) + np.min(grid_1d[:, 1])) / 2.0
    return (centre_y, centre_x)


@decorator_util.jit()
def grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
    mask, pixel_scales, sub_grid_size, origin=(0.0, 0.0)
):
    """ For a sub-grid, every unmasked pixel of a 2D mask array of shape (rows, columns) is divided into a finer \
    uniform grid of shape (sub_grid_size, sub_grid_size). This routine computes the (y,x) arc second coordinates at \
    the centre of every sub-pixel defined by this 2D mask array.

    Coordinates are defined from the top-left corner, where the first unmasked sub-pixel corresponds to index 0. \
    Sub-pixels that are part of the same mask array pixel are indexed next to one another, such that the second \
    sub-pixel in the first pixel has index 1, its next sub-pixel has index 2, and so forth.

    The sub-grid is returned on an array of shape (total_unmasked_pixels*sub_grid_size**2, 2). y coordinates are \
    stored in the 0 index of the second dimension, x coordinates in the 1 index.

    Parameters
     ----------
    mask : ndarray
        A 2D array of bools, where *False* values are unmasked and therefore included as part of the calculated \
        sub-grid.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the 2D mask array.
    sub_grid_size : int
        The size of the sub-grid that each pixel of the 2D mask array is divided into.
    origin : (float, flloat)
        The (y,x) origin of the 2D array, which the sub-grid is shifted around.

    Returns
    --------
    ndarray
        A sub grid of (y,x) arc-second coordinates at the centre of every pixel unmasked pixel on the 2D mask \
        array. The sub grid array has dimensions (total_unmasked_pixels*sub_grid_size**2, 2).

    Examples
    --------
    mask = np.array([[True, False, True],
                     [False, False, False]
                     [True, False, True]])
    grid_1d = sub_grid_1d_from_mask_pixel_scales_and_origin(mask=mask, pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    total_sub_pixels = mask_util.total_sub_pixels_from_mask_and_sub_grid_size(
        mask, sub_grid_size
    )

    grid_1d = np.zeros(shape=(total_sub_pixels, 2))

    centres_arcsec = centres_from_shape_pixel_scales_and_origin(
        shape=mask.shape, pixel_scales=pixel_scales, origin=origin
    )

    sub_index = 0

    y_sub_half = pixel_scales[0] / 2
    y_sub_step = pixel_scales[0] / (sub_grid_size)

    x_sub_half = pixel_scales[1] / 2
    x_sub_step = pixel_scales[1] / (sub_grid_size)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            if not mask[y, x]:

                y_arcsec = (y - centres_arcsec[0]) * pixel_scales[0]
                x_arcsec = (x - centres_arcsec[1]) * pixel_scales[1]

                for y1 in range(sub_grid_size):
                    for x1 in range(sub_grid_size):

                        grid_1d[sub_index, 0] = -(
                            y_arcsec - y_sub_half + y1 * y_sub_step + (y_sub_step / 2.0)
                        )
                        grid_1d[sub_index, 1] = (
                            x_arcsec - x_sub_half + x1 * x_sub_step + (x_sub_step / 2.0)
                        )
                        sub_index += 1

    return grid_1d


def grid_2d_from_mask_pixel_scales_sub_grid_size_and_origin(
    mask, pixel_scales, sub_grid_size, origin=(0.0, 0.0)
):
    """ For a sub-grid, every unmasked pixel of a 2D mask array of shape (rows, columns) is divided into a finer \
    uniform grid of shape (sub_grid_size, sub_grid_size). This routine computes the (y,x) arc second coordinates at \
    the centre of every sub-pixel defined by this 2D mask array.

    Coordinates are defined from the top-left corner, where the first unmasked sub-pixel corresponds to index 0. \
    Sub-pixels that are part of the same mask array pixel are indexed next to one another, such that the second \
    sub-pixel in the first pixel has index 1, its next sub-pixel has index 2, and so forth.

    The sub-grid is returned on an array of shape (total_unmasked_pixels*sub_grid_size**2, 2). y coordinates are \
    stored in the 0 index of the second dimension, x coordinates in the 1 index.

    Parameters
     ----------
    mask : ndarray
        A 2D array of bools, where *False* values are unmasked and therefore included as part of the calculated \
        sub-grid.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the 2D mask array.
    sub_grid_size : int
        The size of the sub-grid that each pixel of the 2D mask array is divided into.
    origin : (float, flloat)
        The (y,x) origin of the 2D array, which the sub-grid is shifted around.

    Returns
    --------
    ndarray
        A sub grid of (y,x) arc-second coordinates at the centre of every pixel unmasked pixel on the 2D mask \
        array. The sub grid array has dimensions (total_unmasked_pixels*sub_grid_size**2, 2).

    Examples
    --------
    mask = np.array([[True, False, True],
                     [False, False, False]
                     [True, False, True]])
    sub_grid_1d = sub_grid_1d_from_mask_pixel_scales_and_origin(mask=mask, pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    grid_1d = grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
        mask=mask, pixel_scales=pixel_scales, sub_grid_size=sub_grid_size, origin=origin
    )

    return grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
        sub_grid_1d=grid_1d, mask=mask, sub_grid_size=sub_grid_size
    )


def grid_1d_from_shape_pixel_scales_sub_grid_size_and_origin(
    shape, pixel_scales, sub_grid_size, origin=(0.0, 0.0)
):
    """ For a sub-grid, every unmasked pixel of a 2D mask array of shape (rows, columns) is divided into a finer \
    uniform grid of shape (sub_grid_size, sub_grid_size). This routine computes the (y,x) arc second coordinates at \
    the centre of every sub-pixel defined by this 2D mask array.

    Coordinates are defined from the top-left corner, where the first sub-pixel corresponds to index [0,0]. \
    Sub-pixels that are part of the same mask array pixel are indexed next to one another, such that the second \
    sub-pixel in the first pixel has index 1, its next sub-pixel has index 2, and so forth.

    The sub-grid is returned on an array of shape (total_pixels**2*sub_grid_size**2, 2). y coordinates are \
    stored in the 0 index of the second dimension, x coordinates in the 1 index.

    Parameters
     ----------
    shape : (int, int)
        The (y,x) shape of the 2D array the sub-grid of coordinates is computed for.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the 2D mask array.
    sub_grid_size : int
        The size of the sub-grid that each pixel of the 2D mask array is divided into.
    origin : (float, flloat)
        The (y,x) origin of the 2D array, which the sub-grid is shifted around.

    Returns
    --------
    ndarray
        A sub grid of (y,x) arc-second coordinates at the centre of every pixel unmasked pixel on the 2D mask \
        array. The sub grid array has dimensions (total_unmasked_pixels*sub_grid_size**2, 2).

    Examples
    --------
    mask = np.array([[True, False, True],
                     [False, False, False]
                     [True, False, True]])
    sub_grid_1d = sub_grid_1d_from_mask_pixel_scales_and_origin(mask=mask, pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """
    return grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
        mask=np.full(fill_value=False, shape=shape),
        pixel_scales=pixel_scales,
        sub_grid_size=sub_grid_size,
        origin=origin,
    )


def grid_2d_from_shape_pixel_scales_sub_grid_size_and_origin(
    shape, pixel_scales, sub_grid_size, origin=(0.0, 0.0)
):
    """ For a sub-grid, every unmasked pixel of a 2D mask array of shape (rows, columns) is divided into a finer \
    uniform grid of shape (sub_grid_size, sub_grid_size). This routine computes the (y,x) arc second coordinates at \
    the centre of every sub-pixel defined by this 2D mask array.

    Coordinates are defined from the top-left corner, where the first sub-pixel corresponds to index [0,0]. \
    Sub-pixels that are part of the same mask array pixel are indexed next to one another, such that the second \
    sub-pixel in the first pixel has index 1, its next sub-pixel has index 2, and so forth.

    The sub-grid is returned on an array of shape (total_pixels**2*sub_grid_size**2, 2). y coordinates are \
    stored in the 0 index of the second dimension, x coordinates in the 1 index.

    Parameters
     ----------
    shape : (int, int)
        The (y,x) shape of the 2D array the sub-grid of coordinates is computed for.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the 2D mask array.
    sub_grid_size : int
        The size of the sub-grid that each pixel of the 2D mask array is divided into.
    origin : (float, flloat)
        The (y,x) origin of the 2D array, which the sub-grid is shifted around.

    Returns
    --------
    ndarray
        A sub grid of (y,x) arc-second coordinates at the centre of every pixel unmasked pixel on the 2D mask \
        array. The sub grid array has dimensions (total_unmasked_pixels*sub_grid_size**2, 2).

    Examples
    --------
    mask = np.array([[True, False, True],
                     [False, False, False]
                     [True, False, True]])
    sub_grid_1d = sub_grid_1d_from_mask_pixel_scales_and_origin(mask=mask, pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """
    return grid_2d_from_mask_pixel_scales_sub_grid_size_and_origin(
        mask=np.full(fill_value=False, shape=shape),
        pixel_scales=pixel_scales,
        sub_grid_size=sub_grid_size,
        origin=origin,
    )


@decorator_util.jit()
def grid_arcsec_1d_to_grid_pixels_1d(
    grid_arcsec_1d, shape, pixel_scales, origin=(0.0, 0.0)
):
    """ Convert a 1D grid of (y,x) arc second coordinates to a 1D grid of (y,x) pixel coordinate values. Pixel
    coordinates are returned as floats such that they include the decimal offset from each pixel's top-left corner
    relative to the input arc-second coordinate.

    The pixel coordinate origin is at the top left corner of the grid, such that the pixel [0,0] corresponds to \
    the highest (most positive) y arc-second coordinate and lowest (most negative) x arc-second coordinate on the gird.

    The arc-second grid is defined by an origin and coordinates are shifted to this origin before computing their \
    1D grid pixel coordinate values.

    The input and output grids are both of shape (total_pixels, 2).

    Parameters
    ----------
    grid_arcsec_1d: ndarray
        The grid of (y,x) coordinates in arc seconds which is converted to pixel value coordinates.
    shape : (int, int)
        The (y,x) shape of the original 2D array the arc-second coordinates were computed on.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the original 2D array.
    origin : (float, flloat)
        The (y,x) origin of the grid, which the arc-second grid is shifted to.

    Returns
    --------
    ndarray
        A grid of (y,x) pixel-value coordinates with dimensions (total_pixels, 2).

    Examples
    --------
    grid_arcsec_1d = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [4.0, 4.0]])
    grid_pixels_1d = grid_arcsec_1d_to_grid_pixels_1d(grid_arcsec_1d=grid_arcsec_1d, shape=(2,2),
                                                           pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    grid_pixels_1d = np.zeros((grid_arcsec_1d.shape[0], 2))

    centres_arcsec = centres_from_shape_pixel_scales_and_origin(
        shape=shape, pixel_scales=pixel_scales, origin=origin
    )

    for i in range(grid_arcsec_1d.shape[0]):

        grid_pixels_1d[i, 0] = (
            (-grid_arcsec_1d[i, 0] / pixel_scales[0]) + centres_arcsec[0] + 0.5
        )
        grid_pixels_1d[i, 1] = (
            (grid_arcsec_1d[i, 1] / pixel_scales[1]) + centres_arcsec[1] + 0.5
        )

    return grid_pixels_1d


@decorator_util.jit()
def grid_arcsec_1d_to_grid_pixel_centres_1d(
    grid_arcsec_1d, shape, pixel_scales, origin=(0.0, 0.0)
):
    """ Convert a 1D grid of (y,x) arc second coordinates to a 1D grid of (y,x) pixel values. Pixel coordinates are \
    returned as integers such that they map directly to the pixel they are contained within.

    The pixel coordinate origin is at the top left corner of the grid, such that the pixel [0,0] corresponds to \
    the highest (most positive) y arc-second coordinate and lowest (most negative) x arc-second coordinate on the gird.

    The arc-second coordinate grid is defined by the class attribute origin, and coordinates are shifted to this \
    origin before computing their 1D grid pixel indexes.

    The input and output grids are both of shape (total_pixels, 2).

    Parameters
    ----------
    grid_arcsec_1d: ndarray
        The grid of (y,x) coordinates in arc seconds which is converted to pixel indexes.
    shape : (int, int)
        The (y,x) shape of the original 2D array the arc-second coordinates were computed on.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the original 2D array.
    origin : (float, flloat)
        The (y,x) origin of the grid, which the arc-second grid is shifted

    Returns
    --------
    ndarray
        A grid of (y,x) pixel indexes with dimensions (total_pixels, 2).

    Examples
    --------
    grid_arcsec_1d = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [4.0, 4.0]])
    grid_pixels_1d = grid_arcsec_1d_to_grid_pixel_centres_1d(grid_arcsec_1d=grid_arcsec_1d, shape=(2,2),
                                                           pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    grid_pixels_1d = np.zeros((grid_arcsec_1d.shape[0], 2))

    centres_arcsec = centres_from_shape_pixel_scales_and_origin(
        shape=shape, pixel_scales=pixel_scales, origin=origin
    )

    for i in range(grid_arcsec_1d.shape[0]):

        grid_pixels_1d[i, 0] = int(
            (-grid_arcsec_1d[i, 0] / pixel_scales[0]) + centres_arcsec[0] + 0.5
        )
        grid_pixels_1d[i, 1] = int(
            (grid_arcsec_1d[i, 1] / pixel_scales[1]) + centres_arcsec[1] + 0.5
        )

    return grid_pixels_1d


@decorator_util.jit()
def grid_arcsec_1d_to_grid_pixel_indexes_1d(
    grid_arcsec_1d, shape, pixel_scales, origin=(0.0, 0.0)
):
    """ Convert a 1D grid of (y,x) arc second coordinates to a 1D grid of (y,x) pixel 1D indexes. Pixel coordinates
    are returned as integers such that they are the pixel from the top-left of the 2D grid going rights and then \
    downwards.

    For example:

    The pixel at the top-left, whose 2D index is [0,0], corresponds to 1D index 0.
    The fifth pixel on the top row, whose 2D index is [0,5], corresponds to 1D index 4.
    The first pixel on the second row, whose 2D index is [0,1], has 1D index 10 if a row has 10 pixels.

    The arc-second coordinate grid is defined by the class attribute origin, and coordinates are shifted to this \
    origin before computing their 1D grid pixel indexes.

    The input and output grids are both of shape (total_pixels, 2).

    Parameters
    ----------
    grid_arcsec_1d: ndarray
        The grid of (y,x) coordinates in arc seconds which is converted to 1D pixel indexes.
    shape : (int, int)
        The (y,x) shape of the original 2D array the arc-second coordinates were computed on.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the original 2D array.
    origin : (float, flloat)
        The (y,x) origin of the grid, which the arc-second grid is shifted.

    Returns
    --------
    ndarray
        A grid of 1d pixel indexes with dimensions (total_pixels, 2).

    Examples
    --------
    grid_arcsec_1d = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [4.0, 4.0]])
    grid_pixels_1d = grid_arcsec_1d_to_grid_pixel_indexes_1d(grid_arcsec_1d=grid_arcsec_1d, shape=(2,2),
                                                           pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    grid_pixels_1d = grid_arcsec_1d_to_grid_pixel_centres_1d(
        grid_arcsec_1d=grid_arcsec_1d,
        shape=shape,
        pixel_scales=pixel_scales,
        origin=origin,
    )

    grid_pixel_indexes_1d = np.zeros(grid_pixels_1d.shape[0])

    for i in range(grid_pixels_1d.shape[0]):

        grid_pixel_indexes_1d[i] = int(
            grid_pixels_1d[i, 0] * shape[1] + grid_pixels_1d[i, 1]
        )

    return grid_pixel_indexes_1d


@decorator_util.jit()
def grid_pixels_1d_to_grid_arcsec_1d(
    grid_pixels_1d, shape, pixel_scales, origin=(0.0, 0.0)
):
    """ Convert a 1D grid of (y,x) pixel coordinates to a 1D grid of (y,x) arc second values.

    The pixel coordinate origin is at the top left corner of the grid, such that the pixel [0,0] corresponds to \
    the highest (most positive) y arc-second coordinate and lowest (most negative) x arc-second coordinate on the gird.

    The arc-second coordinate origin is defined by the class attribute origin, and coordinates are shifted to this \
    origin after computing their values from the 1D grid pixel indexes.

    The input and output grids are both of shape (total_pixels, 2).

    Parameters
    ----------
    grid_pixels_1d: ndarray
        The grid of (y,x) coordinates in pixel values which is converted to arc-second coordinates.
    shape : (int, int)
        The (y,x) shape of the original 2D array the arc-second coordinates were computed on.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the original 2D array.
    origin : (float, flloat)
        The (y,x) origin of the grid, which the arc-second grid is shifted.

    Returns
    --------
    ndarray
        A grid of 1d arc-second coordinates with dimensions (total_pixels, 2).

    Examples
    --------
    grid_pixels_1d = np.array([[0,0], [0,1], [1,0], [1,1])
    grid_pixels_1d = grid_pixels_1d_to_grid_arcsec_1d(grid_pixels_1d=grid_pixels_1d, shape=(2,2),
                                                           pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    grid_arcsec_1d = np.zeros((grid_pixels_1d.shape[0], 2))

    centres_arcsec = centres_from_shape_pixel_scales_and_origin(
        shape=shape, pixel_scales=pixel_scales, origin=origin
    )

    for i in range(grid_arcsec_1d.shape[0]):

        grid_arcsec_1d[i, 0] = (
            -(grid_pixels_1d[i, 0] - centres_arcsec[0] - 0.5) * pixel_scales[0]
        )
        grid_arcsec_1d[i, 1] = (
            grid_pixels_1d[i, 1] - centres_arcsec[1] - 0.5
        ) * pixel_scales[1]

    return grid_arcsec_1d


@decorator_util.jit()
def grid_arcsec_2d_to_grid_pixel_centres_2d(
    grid_arcsec_2d, shape, pixel_scales, origin=(0.0, 0.0)
):
    """ Convert a 2D grid of (y,x) arc second coordinates to a 2D grid of (y,x) pixel values. Pixel coordinates are \
    returned as integers such that they map directly to the pixel they are contained within.

    The pixel coordinate origin is at the top left corner of the grid, such that the pixel [0,0] corresponds to \
    the highest (most positive) y arc-second coordinate and lowest (most negative) x arc-second coordinate on the gird.

    The arc-second coordinate grid is defined by the class attribute origin, and coordinates are shifted to this \
    origin before computing their 1D grid pixel indexes.

    The input and output grids are both of shape (y_dimension, x_dimension, 2).

    Parameters
    ----------
    grid_arcsec_1d: ndarray
        The grid of (y,x) coordinates in arc seconds which is converted to pixel indexes.
    shape : (int, int)
        The (y,x) shape of the original 2D array the arc-second coordinates were computed on.
    pixel_scales : (float, float)
        The (y,x) arc-second to pixel scales of the original 2D array.
    origin : (float, flloat)
        The (y,x) origin of the grid, which the arc-second grid is shifted

    Returns
    --------
    ndarray
        A grid of (y,x) pixel indexes with dimensions (total_pixels, 2).

    Examples
    --------
    grid_arcsec_1d = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [4.0, 4.0]])
    grid_pixels_1d = grid_arcsec_1d_to_grid_pixel_centres_1d(grid_arcsec_1d=grid_arcsec_1d, shape=(2,2),
                                                           pixel_scales=(0.5, 0.5), origin=(0.0, 0.0))
    """

    grid_pixels_2d = np.zeros((grid_arcsec_2d.shape[0], grid_arcsec_2d.shape[1], 2))

    centres_arcsec = centres_from_shape_pixel_scales_and_origin(
        shape=shape, pixel_scales=pixel_scales, origin=origin
    )

    for y in range(grid_arcsec_2d.shape[0]):
        for x in range(grid_arcsec_2d.shape[1]):
            grid_pixels_2d[y, x, 0] = int(
                (-grid_arcsec_2d[y, x, 0] / pixel_scales[0]) + centres_arcsec[0] + 0.5
            )
            grid_pixels_2d[y, x, 1] = int(
                (grid_arcsec_2d[y, x, 1] / pixel_scales[1]) + centres_arcsec[1] + 0.5
            )

    return grid_pixels_2d


@decorator_util.jit()
def furthest_grid_1d_index_from_grid_1d_indexes_and_centre(
    grid_1d, grid_1d_indexes, coordinate
):

    distance_from_centre = 0.0

    for grid_1d_index in grid_1d_indexes:
        y = grid_1d[grid_1d_index, 0]
        x = grid_1d[grid_1d_index, 1]
        distance_from_centre_new = (x - coordinate[1]) ** 2 + (y - coordinate[0]) ** 2
        if distance_from_centre_new >= distance_from_centre:
            distance_from_centre = distance_from_centre_new
            furthest_grid_1d_index = grid_1d_index

    return furthest_grid_1d_index



import autofit as af
import autolens as al
from autolens.pipeline import pipeline

# Up to now, we have passed the priors between phases using the 'customize priors' function. This works nicely, and
# gives us a lot of control for how the prior on every individual parameter is specified. However, it also makes
# the pipeline code longer than it needs to be, and it does not check for typos or errors in the prior linking.

# In this pipeline, we'll pass priors in a slight different way, where the results of a phase are directly passed to
# the next phase. This does not change the behaviour of the pipeline from the previous pipelines, but as you'll see
# reduces the amount of code.

# All template pipelines found in the autolens_workspace use this method of prior passing, so its worth you learning!


def make_pipeline(phase_folders=None):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below. However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline__complex_source"

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/pipeline_tag/phase_name/phase_tag/'
    phase_folders.append(pipeline_name)

    # To begin, we need to initialize the lens's mass model. We should be able to do this by using a simple source
    # model. It won't fit the complicated structure of the source, but it'll give us a reasonable estimate of the
    # einstein radius and the other lens-mass parameters.

    # This should run fine without any prior-passes. In general, a thick, giant ring of source light is something we
    # can be confident MultiNest will fit without much issue, especially when the lens galaxy's light isn't included
    # such that the parameter space is just 12 parameters.

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            source=al.GalaxyModel(
                redshift=1.0, light_0=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.5

    # Now lets add another source component, using the previous model as the initialization on the lens / source
    # parameters. We'll vary the parameters of the lens mass model and first source galaxy component during the fit.

    # To set up phase 2, we use the new method of passing priors, by directly passing the results of phase 1 to the
    # appropirate model components. The 'variable' behaves exactly as it did in the 'customize_priors' function. Hopefully
    # you'll agree the code below is a lot more concise than using the customize_priors functioon!

    phase2 = al.PhaseImaging(
        phase_name="phase_2__lens_sie__source_x2_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase1.result.variable.galaxies.lens
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                light_0=phase1.result.variable.galaxies.source.light_0,
                light_1=al.light_profiles.EllipticalSersic,
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.5

    # Now lets do the same again, but with 3 source galaxy components.

    phase3 = al.PhaseImaging(
        phase_name="phase_3__lens_sie__source_x3_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase2.result.variable.galaxies.lens
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                light_0=phase2.result.variable.galaxies.source.light_0,
                light_1=phase2.result.variable.galaxies.source.light_1,
                light_2=al.light_profiles.EllipticalSersic,
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.5

    # And one more for luck!

    phase4 = al.PhaseImaging(
        phase_name="phase_4__lens_sie__source_x4_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase3.result.variable.galaxies.lens
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                light_0=phase3.result.variable.galaxies.source.light_0,
                light_1=phase3.result.variable.galaxies.source.light_1,
                light_2=phase3.result.variable.galaxies.source.light_2,
                light_3=al.light_profiles.EllipticalSersic,
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase4.optimizer.const_efficiency_mode = True
    phase4.optimizer.n_live_points = 50
    phase4.optimizer.sampling_efficiency = 0.5

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3, phase4)



import autofit as af
import autolens as al

# In this pipeline, we'll perform a basic analysis which initialize a lens model (the lens's light, mass and source's \
# light) and then fits the source galaxy using an inversion. This pipeline uses three phases:

# Phase 1:

# Description: initialize and subtracts the lens light model.
# Lens Light: EllipticalSersic
# Lens Mass: None
# Source Light: None
# Previous Pipelines: None
# Prior Passing: None
# Notes: None

# Phase 2:

# Description: initialize the lens mass model and source light profile, using the lens subtracted image from phase 1.
# Lens Light: None
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Uses the lens light subtracted image from phase 1

# Phase 3:

# Description: Refine the lens light and mass models and source light profile, using priors from the previous 2 phases.
# Lens Light: EllipticalSersic
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens light (variable -> phase 1), lens mass and source (variable -> phase 2)
# Notes: None


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    pixel_scale_interpolation_grid=None,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_init_hyper__lens_sersic_sie__source_sersic"
    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings(
        hyper_galaxies=pipeline_settings.hyper_galaxies,
        hyper_image_sky=pipeline_settings.hyper_image_sky,
        hyper_background_noise=pipeline_settings.hyper_background_noise,
        include_shear=pipeline_settings.include_shear,
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### SETUP SHEAR ###

    # If the pipeline should include shear, add this class below so that it enters the phase.

    # After this pipeline this shear class is passed to all subsequent pipelines, such that the shear is either
    # included or omitted throughout the entire pipeline.

    if pipeline_settings.include_shear:
        shear = al.mass_profiles.ExternalShear
    else:
        shear = None

    ### PHASE 1 ###

    # In phase 1, we will fit only the lens galaxy's light, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.
    # 2) Use a circular mask which includes the lens and source galaxy light.

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens, light=al.light_profiles.EllipticalSersic
            )
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        optimizer_class=af.MultiNest,
    )

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster (if you haven't already, checkout the tutorial '' in howtolens/chapter_2).

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.3

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=pipeline_settings.hyper_galaxies,
        include_background_sky=pipeline_settings.hyper_image_sky,
        include_background_noise=pipeline_settings.hyper_background_noise,
    )

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and source galaxy's light, where we:

    # 1) Use a lens-subtracted image generated by subtracting model lens galaxy image from phase 1.
    # 2) Initialize the priors on the centre of the lens galaxy's mass-profile by linking them to those inferred for \
    #    its light profile in phase 1.
    # 3) Have the option to use an annular mask removing the central light, if the inner_mask_radii parametr is input.

    class LensSubtractedPhase(al.PhaseImaging):
        def customize_priors(self, results):

            ## Lens Mass, Move centre priors to centre of lens light ###

            self.galaxies.lens.mass.centre = (
                results.from_phase("phase_1__lens_sersic")
                .variable_absolute(a=0.1)
                .galaxies.lens.light.centre
            )

    phase2 = LensSubtractedPhase(
        phase_name="phase_2__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                light=phase1.result.constant.galaxies.lens.light,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=shear,
                hyper_galaxy=phase1.result.hyper_combined.constant.galaxies.lens.hyper_galaxy,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source, light=al.light_profiles.EllipticalSersic
            ),
        ),
        hyper_image_sky=phase1.result.hyper_combined.constant.hyper_image_sky,
        hyper_background_noise=phase1.result.hyper_combined.constant.hyper_background_noise,
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = False
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.5

    phase2 = phase2.extend_with_multiple_hyper_phases(
        hyper_galaxy=pipeline_settings.hyper_galaxies,
        include_background_sky=pipeline_settings.hyper_image_sky,
        include_background_noise=pipeline_settings.hyper_background_noise,
    )

    ### PHASE 3 ###

    # In phase 3, we will fit simultaneously the lens and source galaxies, where we:

    # 1) Initialize the lens's light, mass, shear and source's light using the results of phases 1 and 2.
    # 2) Use a circular mask, to fully capture the lens and source light.

    phase3 = al.PhaseImaging(
        phase_name="phase_3__lens_sersic_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                light=phase1.result.variable.galaxies.lens.light,
                mass=phase2.result.variable.galaxies.lens.mass,
                shear=phase2.result.variable.galaxies.lens.shear,
                hyper_galaxy=phase1.result.hyper_combined.constant.galaxies.lens.hyper_galaxy,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source,
                light=phase1.result.variable.galaxies.source.light,
            ),
        ),
        hyper_image_sky=phase2.result.hyper_combined.constant.hyper_image_sky,
        hyper_background_noise=phase2.result.hyper_combined.constant.hyper_background_noise,
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 75
    phase3.optimizer.sampling_efficiency = 0.3

    phase3 = phase3.extend_with_multiple_hyper_phases(
        hyper_galaxy=pipeline_settings.hyper_galaxies,
        include_background_sky=pipeline_settings.hyper_image_sky,
        include_background_noise=pipeline_settings.hyper_background_noise,
    )

    return al.PipelineImaging(pipeline_name, phase1, phase2, phase3, hyper_mode=True)



from autolens import decorator_util
from autolens.data import fourier_transform as ft

import numpy as np
import time

repeats = 1

visibilities = 100000000
image_pixels = 100000
source_pixels = 1000

shape_data = 8 * visibilities
shape_preloads = visibilities * image_pixels * 2
shape_mapping_matrix = visibilities * source_pixels

total_shape = shape_data + shape_preloads + shape_mapping_matrix

print("Data Memory Use (GB) = " + str(shape_data * 8e-9))
print("PreLoad Memory Use (GB) = " + str(shape_preloads * 8e-9))
print("Mapping Matrix Memory Use (GB) = " + str(shape_mapping_matrix * 8e-9))
print("Total Memory Use (GB) = " + str(total_shape * 8e-9))
print()

uv_wavelengths = np.ones(shape=(visibilities, 2))
grid_radians = np.ones(shape=(image_pixels, 2))
image_1d = np.ones(shape=(image_pixels))

transformer = ft.Transformer(
    uv_wavelengths=uv_wavelengths, grid_radians=grid_radians, preload_transform=False
)


@decorator_util.jit()
def interpolation_loop(image_pixels, visibilities):

    for im in range(image_pixels):
        for vis in range(visibilities):
            weight = np.cos(im * vis / 1000000.0) + np.sin(im ** vis)


start = time.time()
for i in range(repeats):
    interpolation_loop(image_pixels=image_pixels, visibilities=visibilities)
diff = time.time() - start
print("Interpolation Loop Time = {}".format(diff / repeats))

stop

start = time.time()
for i in range(repeats):
    transformer.real_visibilities_from_image_1d(image_1d=image_1d)
diff = time.time() - start
print("Real Visibilities Time = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    transformer.real_visibilities_via_preload_from_image(image_1d=image_1d)
diff = time.time() - start
print("Real Visibilities PreLoad Time = {}".format(diff / repeats))



import os

import numpy as np
import pytest

import autolens as al
from autolens import exc

test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


@pytest.fixture(name="array_grid")
def make_array_grid():
    return al.ScaledSquarePixelArray(np.zeros((5, 5)), pixel_scale=0.5)


@pytest.fixture(name="array_grid_rectangular")
def make_array_grid_rectangular():
    return al.ScaledRectangularPixelArray(np.zeros((5, 5)), pixel_scales=(1.0, 0.5))


class TestArrayGeometry:
    class TestArrayAndTuples:
        def test__square_pixel_array__input_data_grid_3x3__centre_is_origin(self):
            data_grid = al.ScaledSquarePixelArray(
                array=np.ones((3, 3)), pixel_scale=1.0
            )

            assert data_grid.pixel_scale == 1.0
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((3.0, 3.0))
            assert data_grid.arc_second_maxima == (1.5, 1.5)
            assert data_grid.arc_second_minima == (-1.5, -1.5)
            assert (data_grid == np.ones((3, 3))).all()

        def test__square_pixel_array__input_data_grid_rectangular__change_origin(self):
            data_grid = al.ScaledSquarePixelArray(
                array=np.ones((4, 3)), pixel_scale=0.1, origin=(1.0, 1.0)
            )

            assert (data_grid == np.ones((4, 3))).all()
            assert data_grid.pixel_scale == 0.1
            assert data_grid.shape == (4, 3)
            assert data_grid.central_pixel_coordinates == (1.5, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((0.4, 0.3))
            assert data_grid.arc_second_maxima == pytest.approx((1.2, 1.15), 1e-4)
            assert data_grid.arc_second_minima == pytest.approx((0.8, 0.85), 1e-4)

            data_grid = al.ScaledSquarePixelArray(
                array=np.ones((3, 4)), pixel_scale=0.1
            )

            assert (data_grid == np.ones((3, 4))).all()
            assert data_grid.pixel_scale == 0.1
            assert data_grid.shape == (3, 4)
            assert data_grid.central_pixel_coordinates == (1.0, 1.5)
            assert data_grid.shape_arcsec == pytest.approx((0.3, 0.4))
            assert data_grid.arc_second_maxima == pytest.approx((0.15, 0.2), 1e-4)
            assert data_grid.arc_second_minima == pytest.approx((-0.15, -0.2), 1e-4)

        def test__rectangular_pixel_grid__input_data_grid_3x3(self):
            data_grid = al.ScaledRectangularPixelArray(
                array=np.ones((3, 3)), pixel_scales=(2.0, 1.0)
            )

            assert data_grid == pytest.approx(np.ones((3, 3)), 1e-4)
            assert data_grid.pixel_scales == (2.0, 1.0)
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((6.0, 3.0))
            assert data_grid.arc_second_maxima == pytest.approx((3.0, 1.5), 1e-4)
            assert data_grid.arc_second_minima == pytest.approx((-3.0, -1.5), 1e-4)

        def test__rectangular_pixel_grid__input_data_grid_rectangular(self):
            data_grid = al.ScaledRectangularPixelArray(
                array=np.ones((4, 3)), pixel_scales=(0.2, 0.1)
            )

            assert data_grid == pytest.approx(np.ones((4, 3)), 1e-4)
            assert data_grid.pixel_scales == (0.2, 0.1)
            assert data_grid.shape == (4, 3)
            assert data_grid.central_pixel_coordinates == (1.5, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((0.8, 0.3), 1e-3)
            assert data_grid.arc_second_maxima == pytest.approx((0.4, 0.15), 1e-4)
            assert data_grid.arc_second_minima == pytest.approx((-0.4, -0.15), 1e-4)

            data_grid = al.ScaledRectangularPixelArray(
                array=np.ones((3, 4)), pixel_scales=(0.1, 0.2)
            )

            assert data_grid == pytest.approx(np.ones((3, 4)), 1e-4)
            assert data_grid.pixel_scales == (0.1, 0.2)
            assert data_grid.shape == (3, 4)
            assert data_grid.central_pixel_coordinates == (1.0, 1.5)
            assert data_grid.shape_arcsec == pytest.approx((0.3, 0.8), 1e-3)
            assert data_grid.arc_second_maxima == pytest.approx((0.15, 0.4), 1e-4)
            assert data_grid.arc_second_minima == pytest.approx((-0.15, -0.4), 1e-4)

        def test__rectangular_pixel_array__input_data_grid_3x3__centre_is_yminus1_xminuss2(
            self
        ):
            data_grid = al.ScaledRectangularPixelArray(
                array=np.ones((3, 3)), pixel_scales=(2.0, 1.0), origin=(-1.0, -2.0)
            )

            assert data_grid == pytest.approx(np.ones((3, 3)), 1e-4)
            assert data_grid.pixel_scales == (2.0, 1.0)
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((6.0, 3.0))

            assert data_grid.origin == (-1.0, -2.0)
            assert data_grid.arc_second_maxima == pytest.approx((2.0, -0.5), 1e-4)
            assert data_grid.arc_second_minima == pytest.approx((-4.0, -3.5), 1e-4)

    class TestCentralPixel:
        def test__square_pixel_grid(self):
            grid = al.ScaledSquarePixelArray(np.zeros((3, 3)), pixel_scale=0.1)
            assert grid.central_pixel_coordinates == (1, 1)

            grid = al.ScaledSquarePixelArray(np.zeros((4, 4)), pixel_scale=0.1)
            assert grid.central_pixel_coordinates == (1.5, 1.5)

            grid = al.ScaledSquarePixelArray(
                np.zeros((5, 3)), pixel_scale=0.1, origin=(1.0, 2.0)
            )
            assert grid.central_pixel_coordinates == (2.0, 1.0)

        def test__rectangular_pixel_grid(self):
            grid = al.ScaledRectangularPixelArray(
                np.zeros((3, 3)), pixel_scales=(2.0, 1.0)
            )
            assert grid.central_pixel_coordinates == (1, 1)

            grid = al.ScaledRectangularPixelArray(
                np.zeros((4, 4)), pixel_scales=(2.0, 1.0)
            )
            assert grid.central_pixel_coordinates == (1.5, 1.5)

            grid = al.ScaledRectangularPixelArray(
                np.zeros((5, 3)), pixel_scales=(2.0, 1.0), origin=(1.0, 2.0)
            )
            assert grid.central_pixel_coordinates == (2, 1)

    class TestGrids:
        def test__square_pixel_grid__grid_2d__compare_to_array_util(self):
            grid_2d_util = al.grid_util.grid_2d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7), pixel_scales=(0.56, 0.56), sub_grid_size=1
            )

            sca = al.ScaledSquarePixelArray(array=np.zeros((4, 7)), pixel_scale=0.56)

            assert sca.grid_2d == pytest.approx(grid_2d_util, 1e-4)

        def test__square_pixel_grid__array_3x3__sets_up_arc_second_grid(self):
            sca = al.ScaledSquarePixelArray(array=np.zeros((3, 3)), pixel_scale=1.0)

            assert (
                sca.grid_2d
                == np.array(
                    [
                        [[1.0, -1.0], [1.0, 0.0], [1.0, 1.0]],
                        [[0.0, -1.0], [0.0, 0.0], [0.0, 1.0]],
                        [[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]],
                    ]
                )
            ).all()

        def test__square_pixel_grid__grid_1d__compare_to_array_util(self):
            grid_1d_util = al.grid_util.grid_1d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7), pixel_scales=(0.56, 0.56), sub_grid_size=1
            )

            sca = al.ScaledSquarePixelArray(array=np.zeros((4, 7)), pixel_scale=0.56)

            assert sca.grid_1d == pytest.approx(grid_1d_util, 1e-4)

        def test__square_pixel_grid__nonzero_centres__compure_to_array_util(self):
            grid_2d_util = al.grid_util.grid_2d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7),
                pixel_scales=(0.56, 0.56),
                sub_grid_size=1,
                origin=(1.0, 3.0),
            )

            sca = al.ScaledSquarePixelArray(
                array=np.zeros((4, 7)), pixel_scale=0.56, origin=(1.0, 3.0)
            )

            assert sca.grid_2d == pytest.approx(grid_2d_util, 1e-4)

            grid_1d_util = al.grid_util.grid_1d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7),
                pixel_scales=(0.56, 0.56),
                sub_grid_size=1,
                origin=(-1.0, -4.0),
            )

            sca = al.ScaledSquarePixelArray(
                array=np.zeros((4, 7)), pixel_scale=0.56, origin=(-1.0, -4.0)
            )

            assert sca.grid_1d == pytest.approx(grid_1d_util, 1e-4)

        def test__rectangular_pixel_grid__grid_2d__compare_to_array_util(self):
            grid_2d_util = al.grid_util.grid_2d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7), pixel_scales=(0.8, 0.56), sub_grid_size=1
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((4, 7)), pixel_scales=(0.8, 0.56)
            )

            assert sca.grid_2d == pytest.approx(grid_2d_util, 1e-4)

        def test__rectangular_pixel_grid__array_3x3__sets_up_arcsecond_grid(self):
            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((3, 3)), pixel_scales=(1.0, 2.0)
            )

            assert (
                sca.grid_2d
                == np.array(
                    [
                        [[1.0, -2.0], [1.0, 0.0], [1.0, 2.0]],
                        [[0.0, -2.0], [0.0, 0.0], [0.0, 2.0]],
                        [[-1.0, -2.0], [-1.0, 0.0], [-1.0, 2.0]],
                    ]
                )
            ).all()

        def test__rectangular_pixel_grid__grid_1d__compare_to_array_util(self):
            grid_1d_util = al.grid_util.grid_1d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7), pixel_scales=(0.8, 0.56), sub_grid_size=1
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((4, 7)), pixel_scales=(0.8, 0.56)
            )

            assert sca.grid_1d == pytest.approx(grid_1d_util, 1e-4)

        def test__rectangular_pixel_grid__nonzero_centres__compure_to_array_util(self):
            grid_2d_util = al.grid_util.grid_2d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7),
                pixel_scales=(0.8, 0.56),
                sub_grid_size=1,
                origin=(1.0, 2.0),
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((4, 7)), pixel_scales=(0.8, 0.56), origin=(1.0, 2.0)
            )

            assert sca.grid_2d == pytest.approx(grid_2d_util, 1e-4)

            grid_1d_util = al.grid_util.grid_1d_from_shape_pixel_scales_sub_grid_size_and_origin(
                shape=(4, 7),
                pixel_scales=(0.8, 0.56),
                sub_grid_size=1,
                origin=(-1.0, -4.0),
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((4, 7)), pixel_scales=(0.8, 0.56), origin=(-1.0, -4.0)
            )

            assert sca.grid_1d == pytest.approx(grid_1d_util, 1e-4)

    class TestConversion:
        def test__arc_second_coordinates_to_pixel_coordinates__arcsec_are_pixel_centres(
            self
        ):
            sca = al.ScaledSquarePixelArray(array=np.zeros((2, 2)), pixel_scale=2.0)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.0, -1.0)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.0, 1.0)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-1.0, -1.0)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-1.0, 1.0)
            ) == (1, 1)

            sca = al.ScaledSquarePixelArray(array=np.zeros((3, 3)), pixel_scale=3.0)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.0, -3.0)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.0, 0.0)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.0, 3.0)
            ) == (0, 2)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, -3.0)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, 0.0)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, 3.0)
            ) == (1, 2)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-3.0, -3.0)
            ) == (2, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-3.0, 0.0)
            ) == (2, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-3.0, 3.0)
            ) == (2, 2)

        def test__arc_second_coordinates_to_pixel_coordinates__arcsec_are_pixel_corners(
            self
        ):
            sca = al.ScaledSquarePixelArray(array=np.zeros((2, 2)), pixel_scale=2.0)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.99, -1.99)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.99, -0.01)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.01, -1.99)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.01, -0.01)
            ) == (0, 0)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(2.01, 0.01)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(2.01, 1.99)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.01, 0.01)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.01, 1.99)
            ) == (0, 1)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.01, -1.99)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.01, -0.01)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-1.99, -1.99)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-1.99, -0.01)
            ) == (1, 0)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.01, 0.01)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.01, 1.99)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-1.99, 0.01)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-1.99, 1.99)
            ) == (1, 1)

        def test__arc_second_coordinates_to_pixel_coordinates__arcsec_are_pixel_centres__nonzero_centre(
            self
        ):
            sca = al.ScaledSquarePixelArray(
                array=np.zeros((2, 2)), pixel_scale=2.0, origin=(1.0, 1.0)
            )

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(2.0, 0.0)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(2.0, 2.0)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, 0.0)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, 2.0)
            ) == (1, 1)

            sca = al.ScaledSquarePixelArray(
                array=np.zeros((3, 3)), pixel_scale=3.0, origin=(3.0, 3.0)
            )

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(6.0, 0.0)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(6.0, 3.0)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(6.0, 6.0)
            ) == (0, 2)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.0, 0.0)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.0, 3.0)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.0, 6.0)
            ) == (1, 2)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, 0.0)
            ) == (2, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, 3.0)
            ) == (2, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.0, 6.0)
            ) == (2, 2)

        def test__arc_second_coordinates_to_pixel_coordinates__arcsec_are_pixel_corners__nonzero_centre(
            self
        ):
            sca = al.ScaledSquarePixelArray(
                array=np.zeros((2, 2)), pixel_scale=2.0, origin=(1.0, 1.0)
            )

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(2.99, -0.99)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(2.99, 0.99)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.01, -0.99)
            ) == (0, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.01, 0.99)
            ) == (0, 0)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.01, 1.01)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(3.01, 2.99)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.01, 1.01)
            ) == (0, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(1.01, 2.99)
            ) == (0, 1)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.99, -0.99)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.99, 0.99)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.99, -0.99)
            ) == (1, 0)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.99, 0.99)
            ) == (1, 0)

            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.99, 1.01)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(0.99, 2.99)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.99, 1.01)
            ) == (1, 1)
            assert sca.arc_second_coordinates_to_pixel_coordinates(
                arc_second_coordinates=(-0.99, 2.99)
            ) == (1, 1)

        def test__square_pixel_grid__1d_arc_second_grid_to_1d_pixel_centred_grid__same_as_imaging_util(
            self
        ):
            grid_arcsec = np.array([[0.5, -0.5], [0.5, 0.5], [-0.5, -0.5], [-0.5, 0.5]])

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_centres_1d(
                grid_arcsec_1d=grid_arcsec, shape=(2, 2), pixel_scales=(2.0, 2.0)
            )

            sca = al.ScaledSquarePixelArray(array=np.zeros((2, 2)), pixel_scale=2.0)

            grid_pixels = sca.grid_arcsec_to_grid_pixel_centres(grid_arcsec=grid_arcsec)

            assert (grid_pixels == grid_pixels_util).all()

        def test__square_pixel_grid__1d_arc_second_grid_to_1d_pixel_indexes_grid__same_as_imaging_util(
            self
        ):
            grid_arcsec = np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])

            grid_pixel_indexes_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec_1d=grid_arcsec, shape=(2, 2), pixel_scales=(2.0, 2.0)
            )

            sca = al.ScaledSquarePixelArray(array=np.zeros((2, 2)), pixel_scale=2.0)

            grid_pixel_indexes = sca.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec=grid_arcsec
            )

            assert (grid_pixel_indexes == grid_pixel_indexes_util).all()

        def test__rectangular_pixel_grid__1d_arc_second_grid_to_1d_pixel_centred_grid__same_as_imaging_util(
            self
        ):
            grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_centres_1d(
                grid_arcsec_1d=grid_arcsec, shape=(2, 2), pixel_scales=(7.0, 2.0)
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((2, 2)), pixel_scales=(7.0, 2.0)
            )

            grid_pixels = sca.grid_arcsec_to_grid_pixel_centres(grid_arcsec=grid_arcsec)

            assert (grid_pixels == grid_pixels_util).all()

        def test__rectangular_pixel_grid__1d_arc_second_grid_to_1d_pixel_indexes_grid__same_as_imaging_util(
            self
        ):
            grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec_1d=grid_arcsec, shape=(2, 2), pixel_scales=(2.0, 4.0)
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((2, 2)), pixel_scales=(2.0, 4.0)
            )

            grid_pixels = sca.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec=grid_arcsec
            )

            assert (grid_pixels == grid_pixels_util).all()

        def test__rectangular_pixel_grid__1d_arc_second_grid_to_1d_pixel_grid__same_as_imaging_util(
            self
        ):
            grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixels_1d(
                grid_arcsec_1d=grid_arcsec, shape=(2, 2), pixel_scales=(2.0, 4.0)
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((2, 2)), pixel_scales=(2.0, 4.0)
            )

            grid_pixels = sca.grid_arcsec_to_grid_pixels(grid_arcsec=grid_arcsec)

            assert (grid_pixels == grid_pixels_util).all()

        def test__square_pixel_grid__1d_pixel_grid_to_1d_pixel_centred_grid__same_as_imaging_util(
            self
        ):
            grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

            grid_pixels_util = al.grid_util.grid_pixels_1d_to_grid_arcsec_1d(
                grid_pixels_1d=grid_pixels, shape=(2, 2), pixel_scales=(2.0, 2.0)
            )

            sca = al.ScaledSquarePixelArray(array=np.zeros((2, 2)), pixel_scale=2.0)

            grid_pixels = sca.grid_pixels_to_grid_arcsec(grid_pixels=grid_pixels)

            assert (grid_pixels == grid_pixels_util).all()

        def test__square_pixel_grid__1d_pixel_grid_to_1d_pixel_grid__same_as_imaging_util(
            self
        ):
            grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

            grid_pixels_util = al.grid_util.grid_pixels_1d_to_grid_arcsec_1d(
                grid_pixels_1d=grid_pixels, shape=(2, 2), pixel_scales=(2.0, 2.0)
            )

            sca = al.ScaledSquarePixelArray(array=np.zeros((2, 2)), pixel_scale=2.0)

            grid_pixels = sca.grid_pixels_to_grid_arcsec(grid_pixels=grid_pixels)

            assert (grid_pixels == grid_pixels_util).all()

        def test__square_pixel_grid__grids_with_nonzero_centres__same_as_imaging_util(
            self
        ):
            grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

            sca = al.ScaledSquarePixelArray(
                array=np.zeros((2, 2)), pixel_scale=2.0, origin=(1.0, 2.0)
            )

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixels_1d(
                grid_arcsec_1d=grid_arcsec,
                shape=(2, 2),
                pixel_scales=(2.0, 2.0),
                origin=(1.0, 2.0),
            )
            grid_pixels = sca.grid_arcsec_to_grid_pixels(grid_arcsec=grid_arcsec)
            assert (grid_pixels == grid_pixels_util).all()

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec_1d=grid_arcsec,
                shape=(2, 2),
                pixel_scales=(2.0, 2.0),
                origin=(1.0, 2.0),
            )
            grid_pixels = sca.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec=grid_arcsec
            )
            assert grid_pixels == pytest.approx(grid_pixels_util, 1e-4)

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_centres_1d(
                grid_arcsec_1d=grid_arcsec,
                shape=(2, 2),
                pixel_scales=(2.0, 2.0),
                origin=(1.0, 2.0),
            )
            grid_pixels = sca.grid_arcsec_to_grid_pixel_centres(grid_arcsec=grid_arcsec)
            assert grid_pixels == pytest.approx(grid_pixels_util, 1e-4)

            grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

            grid_arcsec_util = al.grid_util.grid_pixels_1d_to_grid_arcsec_1d(
                grid_pixels_1d=grid_pixels,
                shape=(2, 2),
                pixel_scales=(2.0, 2.0),
                origin=(1.0, 2.0),
            )

            grid_arcsec = sca.grid_pixels_to_grid_arcsec(grid_pixels=grid_pixels)

            assert (grid_arcsec == grid_arcsec_util).all()

        def test__rectangular_pixel_grid__grids_with_nonzero_centres__same_as_imaging_util(
            self
        ):
            grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

            sca = al.ScaledRectangularPixelArray(
                array=np.zeros((2, 2)), pixel_scales=(2.0, 1.0), origin=(1.0, 2.0)
            )

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixels_1d(
                grid_arcsec_1d=grid_arcsec,
                shape=(2, 2),
                pixel_scales=(2.0, 1.0),
                origin=(1.0, 2.0),
            )
            grid_pixels = sca.grid_arcsec_to_grid_pixels(grid_arcsec=grid_arcsec)
            assert (grid_pixels == grid_pixels_util).all()

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec_1d=grid_arcsec,
                shape=(2, 2),
                pixel_scales=(2.0, 1.0),
                origin=(1.0, 2.0),
            )
            grid_pixels = sca.grid_arcsec_1d_to_grid_pixel_indexes_1d(
                grid_arcsec=grid_arcsec
            )
            assert (grid_pixels == grid_pixels_util).all()

            grid_pixels_util = al.grid_util.grid_arcsec_1d_to_grid_pixel_centres_1d(
                grid_arcsec_1d=grid_arcsec,
                shape=(2, 2),
                pixel_scales=(2.0, 1.0),
                origin=(1.0, 2.0),
            )
            grid_pixels = sca.grid_arcsec_to_grid_pixel_centres(grid_arcsec=grid_arcsec)
            assert grid_pixels == pytest.approx(grid_pixels_util, 1e-4)

            grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

            grid_arcsec_util = al.grid_util.grid_pixels_1d_to_grid_arcsec_1d(
                grid_pixels_1d=grid_pixels,
                shape=(2, 2),
                pixel_scales=(2.0, 1.0),
                origin=(1.0, 2.0),
            )

            grid_arcsec = sca.grid_pixels_to_grid_arcsec(grid_pixels=grid_pixels)

            assert (grid_arcsec == grid_arcsec_util).all()

    class TestTicks:
        def test__square_pixel_grid__yticks(self):
            sca = al.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
            assert sca.yticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

            sca = al.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=0.5)
            assert sca.yticks == pytest.approx(
                np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
            )

            sca = al.ScaledSquarePixelArray(array=np.ones((6, 3)), pixel_scale=1.0)
            assert sca.yticks == pytest.approx(np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3)

            sca = al.ScaledSquarePixelArray(array=np.ones((3, 1)), pixel_scale=1.0)
            assert sca.yticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        def test__square_pixel_grid__xticks(self):
            sca = al.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
            assert sca.xticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

            sca = al.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=0.5)
            assert sca.xticks == pytest.approx(
                np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
            )

            sca = al.ScaledSquarePixelArray(array=np.ones((3, 6)), pixel_scale=1.0)
            assert sca.xticks == pytest.approx(np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3)

            sca = al.ScaledSquarePixelArray(array=np.ones((1, 3)), pixel_scale=1.0)
            assert sca.xticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        def test__rectangular_pixel_grid__yticks(self):
            sca = al.ScaledRectangularPixelArray(
                array=np.ones((3, 3)), pixel_scales=(1.0, 5.0)
            )
            assert sca.yticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

            sca = al.ScaledRectangularPixelArray(
                array=np.ones((3, 3)), pixel_scales=(0.5, 5.0)
            )
            assert sca.yticks == pytest.approx(
                np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.ones((6, 3)), pixel_scales=(1.0, 5.0)
            )
            assert sca.yticks == pytest.approx(np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3)

            sca = al.ScaledRectangularPixelArray(
                array=np.ones((3, 6)), pixel_scales=(1.0, 5.0)
            )
            assert sca.yticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        def test__rectangular_pixel_grid__xticks(self):
            sca = al.ScaledRectangularPixelArray(
                array=np.ones((3, 3)), pixel_scales=(5.0, 1.0)
            )
            assert sca.xticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

            sca = al.ScaledRectangularPixelArray(
                array=np.ones((3, 3)), pixel_scales=(5.0, 0.5)
            )
            assert sca.xticks == pytest.approx(
                np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
            )

            sca = al.ScaledRectangularPixelArray(
                array=np.ones((3, 6)), pixel_scales=(5.0, 1.0)
            )
            assert sca.xticks == pytest.approx(np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3)

            sca = al.ScaledRectangularPixelArray(
                array=np.ones((6, 3)), pixel_scales=(5.0, 1.0)
            )
            assert sca.xticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)


class TestArray:
    class TestResizing:
        def test__pad__compare_to_imaging_util(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = al.ScaledSquarePixelArray(array, pixel_scale=1.0)

            modified = array.resized_scaled_array_from_array(
                new_shape=(7, 7), new_centre_pixels=(1, 1)
            )

            modified_util = al.array_util.resized_array_2d_from_array_2d_and_resized_shape(
                array_2d=array, resized_shape=(7, 7), origin=(1, 1)
            )

            assert type(modified) == al.ScaledSquarePixelArray
            assert (modified == modified_util).all()
            assert modified.pixel_scale == 1.0

        def test__trim__compare_to_imaging_util(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = al.ScaledSquarePixelArray(array, pixel_scale=1.0)

            modified = array.resized_scaled_array_from_array(
                new_shape=(3, 3), new_centre_pixels=(4, 4)
            )

            modified_util = al.array_util.resized_array_2d_from_array_2d_and_resized_shape(
                array_2d=array, resized_shape=(3, 3), origin=(4, 4)
            )

            assert type(modified) == al.ScaledSquarePixelArray
            assert (modified == modified_util).all()
            assert modified.pixel_scale == 1.0

        def test__new_centre_is_in_arcsec(self):
            array = np.ones((5, 5))
            array[2, 2] = 2.0

            array = al.ScaledSquarePixelArray(array, pixel_scale=3.0)

            modified = array.resized_scaled_array_from_array(
                new_shape=(3, 3), new_centre_arcsec=(6.0, 6.0)
            )
            modified_util = al.array_util.resized_array_2d_from_array_2d_and_resized_shape(
                array_2d=array, resized_shape=(3, 3), origin=(0, 4)
            )
            assert (modified == modified_util).all()

            modified = array.resized_scaled_array_from_array(
                new_shape=(3, 3), new_centre_arcsec=(7.49, 4.51)
            )
            modified_util = al.array_util.resized_array_2d_from_array_2d_and_resized_shape(
                array_2d=array, resized_shape=(3, 3), origin=(0, 4)
            )
            assert (modified == modified_util).all()

            modified = array.resized_scaled_array_from_array(
                new_shape=(3, 3), new_centre_arcsec=(7.49, 7.49)
            )
            modified_util = al.array_util.resized_array_2d_from_array_2d_and_resized_shape(
                array_2d=array, resized_shape=(3, 3), origin=(0, 4)
            )
            assert (modified == modified_util).all()

            modified = array.resized_scaled_array_from_array(
                new_shape=(3, 3), new_centre_arcsec=(4.51, 4.51)
            )
            modified_util = al.array_util.resized_array_2d_from_array_2d_and_resized_shape(
                array_2d=array, resized_shape=(3, 3), origin=(0, 4)
            )
            assert (modified == modified_util).all()

            modified = array.resized_scaled_array_from_array(
                new_shape=(3, 3), new_centre_arcsec=(4.51, 7.49)
            )
            modified_util = al.array_util.resized_array_2d_from_array_2d_and_resized_shape(
                array_2d=array, resized_shape=(3, 3), origin=(0, 4)
            )
            assert (modified == modified_util).all()


class TestScaledSquarePixelArray:
    class TestConstructors(object):
        def test__constructor(self, array_grid):
            # Does the array grid class correctly instantiate as an instance of ndarray?
            assert array_grid.shape == (5, 5)
            assert array_grid.pixel_scale == 0.5
            assert isinstance(array_grid, np.ndarray)
            assert isinstance(array_grid, al.ScaledSquarePixelArray)

        def test__init__input_data_grid_single_value__all_attributes_correct_including_data_inheritance(
            self
        ):
            data_grid = al.ScaledSquarePixelArray.single_value(
                value=5.0, shape=(3, 3), pixel_scale=1.0, origin=(1.0, 1.0)
            )

            assert (data_grid == 5.0 * np.ones((3, 3))).all()
            assert data_grid.pixel_scale == 1.0
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((3.0, 3.0))
            assert data_grid.origin == (1.0, 1.0)

        def test__from_fits__input_data_grid_3x3__all_attributes_correct_including_data_inheritance(
            self
        ):
            data_grid = al.ScaledSquarePixelArray.from_fits_with_pixel_scale(
                file_path=test_data_dir + "3x3_ones.fits",
                hdu=0,
                pixel_scale=1.0,
                origin=(1.0, 1.0),
            )

            assert (data_grid == np.ones((3, 3))).all()
            assert data_grid.pixel_scale == 1.0
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((3.0, 3.0))
            assert data_grid.origin == (1.0, 1.0)

        def test__from_fits__input_data_grid_4x3__all_attributes_correct_including_data_inheritance(
            self
        ):
            data_grid = al.ScaledSquarePixelArray.from_fits_with_pixel_scale(
                file_path=test_data_dir + "4x3_ones.fits", hdu=0, pixel_scale=0.1
            )

            assert (data_grid == np.ones((4, 3))).all()
            assert data_grid.pixel_scale == 0.1
            assert data_grid.shape == (4, 3)
            assert data_grid.central_pixel_coordinates == (1.5, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((0.4, 0.3))

        def test__zero_or_negative_pixel_scale__raises_exception(self):
            with pytest.raises(exc.ScaledArrayException):
                al.ScaledSquarePixelArray(array=np.ones((2, 2)), pixel_scale=0.0)

            with pytest.raises(exc.ScaledArrayException):
                al.ScaledSquarePixelArray(array=np.ones((2, 2)), pixel_scale=-0.5)

    class TestExtract:
        def test__mask_extract_2d_array__uses_the_limits_of_the_mask(self):
            array = np.array(
                [
                    [1.0, 2.0, 3.0, 4.0],
                    [5.0, 6.0, 7.0, 8.0],
                    [9.0, 10.0, 11.0, 12.0],
                    [13.0, 14.0, 15.0, 16.0],
                ]
            )

            array = al.ScaledSquarePixelArray(array=array, pixel_scale=1.0)

            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            array_extracted = array.zoomed_scaled_array_around_mask(mask=mask, buffer=0)
            assert (array_extracted == np.array([[6.0, 7.0], [10.0, 11.0]])).all()

            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, False, False, False],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            array_extracted = array.zoomed_scaled_array_around_mask(mask=mask, buffer=0)
            assert (
                array_extracted == np.array([[6.0, 7.0, 8.0], [10.0, 11.0, 12.0]])
            ).all()

            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, False, False, True],
                        [True, True, False, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            array_extracted = array.zoomed_scaled_array_around_mask(mask=mask, buffer=0)
            assert (
                array_extracted == np.array([[6.0, 7.0], [10.0, 11.0], [14.0, 15.0]])
            ).all()

            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [False, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            array_extracted = array.zoomed_scaled_array_around_mask(mask=mask, buffer=0)
            assert (
                array_extracted == np.array([[5.0, 6.0, 7.0], [9.0, 10.0, 11.0]])
            ).all()

            mask = al.Mask(
                array=np.array(
                    [
                        [True, False, True, True],
                        [True, False, False, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            array_extracted = array.zoomed_scaled_array_around_mask(mask=mask, buffer=0)
            assert (
                array_extracted == np.array([[2.0, 3.0], [6.0, 7.0], [10.0, 11.0]])
            ).all()

            mask = al.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            array_extracted = array.zoomed_scaled_array_around_mask(mask=mask, buffer=1)
            assert (
                array_extracted
                == np.array(
                    [
                        [1.0, 2.0, 3.0, 4.0],
                        [5.0, 6.0, 7.0, 8.0],
                        [9.0, 10.0, 11.0, 12.0],
                        [13.0, 14.0, 15.0, 16.0],
                    ]
                )
            ).all()

    class TestBinUp:
        def test__compare_all_extract_methods_to_array_util(self):
            array_2d = np.array(
                [
                    [1.0, 6.0, 3.0, 7.0, 3.0, 2.0],
                    [2.0, 5.0, 3.0, 7.0, 7.0, 7.0],
                    [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                ]
            )

            array_2d = al.ScaledSquarePixelArray(array=array_2d, pixel_scale=0.1)

            binned_up_array_2d_util = al.binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
                array_2d=array_2d, bin_up_factor=4
            )
            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=4, method="mean"
            )
            assert (binned_up_array_2d_util == binned_up_scaled_array).all()

            binned_up_array_2d_util = al.binning_util.binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(
                array_2d=array_2d, bin_up_factor=4
            )
            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=4, method="quadrature"
            )
            assert (binned_up_array_2d_util == binned_up_scaled_array).all()

            binned_up_array_2d_util = al.binning_util.binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(
                array_2d=array_2d, bin_up_factor=4
            )
            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=4, method="sum"
            )
            assert (binned_up_array_2d_util == binned_up_scaled_array).all()

        def test__pixel_scale_of_scaled_arrays_are_updated(self):
            array_2d = np.array(
                [
                    [1.0, 6.0, 3.0, 7.0, 3.0, 2.0],
                    [2.0, 5.0, 3.0, 7.0, 7.0, 7.0],
                    [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                ]
            )

            array_2d = al.ScaledSquarePixelArray(array=array_2d, pixel_scale=0.1)

            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=4, method="mean"
            )
            assert binned_up_scaled_array.pixel_scale == pytest.approx(0.4, 1.0e-4)
            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=6, method="mean"
            )
            assert binned_up_scaled_array.pixel_scale == pytest.approx(0.6, 1.0e-4)

            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=4, method="quadrature"
            )
            assert binned_up_scaled_array.pixel_scale == pytest.approx(0.4, 1.0e-4)
            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=6, method="quadrature"
            )
            assert binned_up_scaled_array.pixel_scale == pytest.approx(0.6, 1.0e-4)

            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=4, method="sum"
            )
            assert binned_up_scaled_array.pixel_scale == pytest.approx(0.4, 1.0e-4)
            binned_up_scaled_array = array_2d.binned_up_array_from_array(
                bin_up_factor=6, method="sum"
            )
            assert binned_up_scaled_array.pixel_scale == pytest.approx(0.6, 1.0e-4)

        def test__invalid_method__raises_exception(self):
            array_2d = np.array(
                [
                    [1.0, 6.0, 3.0, 7.0, 3.0, 2.0],
                    [2.0, 5.0, 3.0, 7.0, 7.0, 7.0],
                    [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                ]
            )

            array_2d = al.ScaledSquarePixelArray(array=array_2d, pixel_scale=0.1)
            with pytest.raises(exc.DataException):
                array_2d.binned_up_array_from_array(bin_up_factor=4, method="wrong")


class TestScaledRectangularPixelArray:
    class TestConstructors(object):
        def test__init__input_data_grid_single_value__all_attributes_correct_including_data_inheritance(
            self
        ):
            data_grid = al.ScaledRectangularPixelArray.single_value(
                value=5.0, shape=(3, 3), pixel_scales=(2.0, 1.0), origin=(1.0, 1.0)
            )

            assert (data_grid == 5.0 * np.ones((3, 3))).all()
            assert data_grid.pixel_scales == (2.0, 1.0)
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((6.0, 3.0))
            assert data_grid.origin == (1.0, 1.0)

        def test__from_fits__input_data_grid_3x3__all_attributes_correct_including_data_inheritance(
            self
        ):
            data_grid = al.ScaledRectangularPixelArray.from_fits_with_pixel_scale(
                file_path=test_data_dir + "3x3_ones.fits",
                hdu=0,
                pixel_scales=(2.0, 1.0),
                origin=(1.0, 1.0),
            )

            assert data_grid == pytest.approx(np.ones((3, 3)), 1e-4)
            assert data_grid.pixel_scales == (2.0, 1.0)
            assert data_grid.shape == (3, 3)
            assert data_grid.central_pixel_coordinates == (1.0, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((6.0, 3.0))
            assert data_grid.origin == (1.0, 1.0)

        def test__from_fits__input_data_grid_4x3__all_attributes_correct_including_data_inheritance(
            self
        ):
            data_grid = al.ScaledRectangularPixelArray.from_fits_with_pixel_scale(
                file_path=test_data_dir + "4x3_ones.fits",
                hdu=0,
                pixel_scales=(0.2, 0.1),
            )

            assert data_grid == pytest.approx(np.ones((4, 3)), 1e-4)
            assert data_grid.pixel_scales == (0.2, 0.1)
            assert data_grid.shape == (4, 3)
            assert data_grid.central_pixel_coordinates == (1.5, 1.0)
            assert data_grid.shape_arcsec == pytest.approx((0.8, 0.3))

        def test__zero_or_negative_pixel_scale__raises_exception(self):
            with pytest.raises(exc.ScaledArrayException):
                al.ScaledRectangularPixelArray(
                    array=np.ones((2, 2)), pixel_scales=(0.0, 0.5)
                )

            with pytest.raises(exc.ScaledArrayException):
                al.ScaledRectangularPixelArray(
                    array=np.ones((2, 2)), pixel_scales=(0.5, 0.0)
                )

            with pytest.raises(exc.ScaledArrayException):
                al.ScaledRectangularPixelArray(
                    array=np.ones((2, 2)), pixel_scales=(-0.5, 0.5)
                )

            with pytest.raises(exc.ScaledArrayException):
                al.ScaledRectangularPixelArray(
                    array=np.ones((2, 2)), pixel_scales=(0.5, -0.5)
                )



import autolens as al
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.inversion.plotters import inversion_plotters

# So, we can use an inversion to reconstruct an image. Furthermore, this reconstruction provides the 'best-fit'
# solution. And, when we inspect the fit with the fitting module, we see residuals indicative of a good fit.

# Everything sounds pretty good, doesn't it? You're probably thinking, why are there more tutorials? We can use
# inversions now, don't ruin it! Well, there is a problem - which I hid from you in the last tutorial, which we'll
# cover now.

# Lets use the same simple source as last time.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy_0 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=0.3,
            sersic_index=1.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy_0])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# We're going to perform a lot of fits using an inversion this tutorial. This would create a lot of code, so to keep
# things tidy, I've setup this function which handles it all for us.

# (You may notice we include an option to 'use_inversion_border, ignore this for now, as we'll be covering borders in
# the next tutorial)
def perform_fit_with_source_galaxy(source_galaxy):

    ccd_data = simulate()

    mask = al.Mask.circular_annular(
        shape=ccd_data.shape,
        pixel_scale=ccd_data.pixel_scale,
        inner_radius_arcsec=0.5,
        outer_radius_arcsec=2.2,
    )

    lens_data = al.LensData(ccd_data=ccd_data, mask=mask)

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)


# Okay, so lets look at our fit from the previous tutorial in more detail. We'll use a higher resolution 40 x 40 grid.
source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=1.0),
)

fit = perform_fit_with_source_galaxy(source_galaxy=source_galaxy)

al.lens_fit_plotters.plot_fit_subplot(fit=fit)

# It looks pretty good! However, this is because I sneakily chose a regularization coefficient that gives a good looking
# solution. If we reduce this regularization coefficient to zero, our source reconstruction goes weird.
source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=0.0),
)

no_regularization_fit = perform_fit_with_source_galaxy(source_galaxy=source_galaxy)

al.lens_fit_plotters.plot_fit_subplot(
    fit=no_regularization_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

# So, what's happening here? Why does reducing the regularization do this to our source reconstruction?

# When our inversion reconstructs a source, it doesn't *just* compute the set of fluxes that best-fit the image. It also
# 'regularizes' this solution, going to every pixel on our rectangular grid and comparing its reconstructed flux with
# its 4 neighboring pixels. If the difference in flux is large the solution is penalized, reducing its likelihood.
# You can think of this as us applying a prior that our source galaxy solution is 'smooth'.

# This adds a 'penalty term' to the likelihood of an inversion which is the summed difference between the
# reconstructed fluxes of every source-pixel pair multiplied by the regularization coefficient. By setting the
# regularization coefficient to zero, we set this penalty term to zero, meaning that regularization is omitted.

# Why do we need to regularize our solution? Well, we just saw why - if we don't apply this smoothing, we 'over-fit'
# the image. More specifically, we over-fit the noise in the image, which is what the large flux values located
# at the exteriors of the source reconstruction are doing. Think about it, if your sole aim is to maximize the
# likelihood, the best way to do this is to fit *everything* accurately, including the noise.

# If we change the 'normalization' variables of the plotter such that the color-map is restricted to a narrower
# range of values, we can see that even without regularization we are still reconstructing the actual source galaxy.
inversion_plotters.plot_pixelization_values(
    inversion=no_regularization_fit.inversion, norm_max=1.0, norm_min=-1.0
)

# Over-fitting is why regularization is necessary. Solutions like this completely ruin our attempts to model a
# strong lens. By smoothing our source reconstruction we ensure it doesn't fit the noise in the image. If we set a
# really high regularization coefficient we completely remove over-fitting at the expense of also fitting the
# image less accurately.
source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=100.0),
)

high_regularization_fit = perform_fit_with_source_galaxy(source_galaxy=source_galaxy)

al.lens_fit_plotters.plot_fit_subplot(
    fit=high_regularization_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

# So, we now understand regularization and its purpose. But there is one nagging question that remains, how do I choose
# the regularization coefficient? We can't use our likelihood, as decreasing the regularization coefficient will always
# increase the likelihood, because it allows the source reconstruction to fit the data better.
print("Likelihood Without Regularization:")
print(no_regularization_fit.likelihood_with_regularization)
print("Likelihood With Normal Regularization:")
print(fit.likelihood_with_regularization)
print("Likelihood With High Regularization:")
print(high_regularization_fit.likelihood_with_regularization)

# If we used the likelihood we will always choose a coefficient of 0! We need a different goodness-of-fit measure. For
# this, we invoke the 'Bayesian evidence', which quantifies the goodness of the fit as follows:

# - First, it requires that the residuals of the fit are consistent with Gaussian noise (which is the noise expected
#   in CCD imaging). If this Gaussian pattern is not visible in the residuals, it tells us that the noise must have been
#   over-fitted. Thus, the Bayesian evidence decreases. Obviously, if the image is poorly fitted, the residuals don't
#   appear Gaussian either, but the poor fit will lead to a decrease in Bayesian evidence decreases all the same!

# - This leaves us with a large number of solutions which all fit the data equally well (e.g., to the noise level). To
#   determine the best-fit from these solutions the Bayesian evidence quantifies the complexity of each solution's
#   source reconstruction. If the inversion requires lots of pixels and a low level of regularization to achieve a good
#   fit, the Bayesian evidence decreases. It penalizes solutions which are complex, which, in a Bayesian sense, are less
#   probable (you may want to look up 'Occam's Razor').

# If a really complex source reconstruction is paramount to fitting the image accurately than that is probably the
# correct solution. However, the Bayesian evidence ensures we only invoke this more complex solution when the data
# necessitates it.

# Lets take a look at the Bayesian evidence:
print("Bayesian Evidence Without Regularization:")
print(no_regularization_fit.evidence)
print("Bayesian Evidence With Normal Regularization:")
print(fit.evidence)
print("Bayesian Evidence With High Regularization:")
print(high_regularization_fit.evidence)

# Great! As expected, the solution that we could see 'by-eye' was the best solution corresponds to
# the highest evidence solution.

# Before we end, lets consider which aspects of an inversion are linear and which are non-linear.

# The linear part of the linear inversion solves for the 'best-fit' solution. For a given regularizaton coefficient,
# this includes the regularization pattern. That is, we linearly reconstruct the combination of source-pixel fluxes
# that best-fit the image *including* the penalty term due to comparing neighboring source-pixel fluxes.

# However, determining the regularization coefficient that maximizes the Bayesian evidence remains a non-linear
# problem and this becomes part of our non-linear search. The Bayesian evidence also depends on the source
# resolution which means the pixel-grid resolution may also be part of our non-linear search. Nevertheless,
# this is only 3 parameters - there were 30+ when using light profiles to represent the source!

# Here are a few questions for you to think about.

# 1) We maximize the evidence by using simpler source reconstructions. Therefore, decreasing the pixel-grid size should
#    provide a higher evidence, provided it still has enough resolution to fit the image well (and provided that the
#    regularization coefficient is still an appropriate value). Can you increase the evidence from the value above by
#    changing these parameters - I've set you up with a code to do so below.
source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=1.0),
)

fit = perform_fit_with_source_galaxy(source_galaxy=source_galaxy)

print("Previous Bayesian Evidence:")
print(10395.370224426646)
print("New Bayesian Evidence:")
print(fit.evidence)

al.lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# 2) Can you think of any other ways we might increase the evidence even further? If not - don't worry about - but
#    you'll learn that PyAutoLens actually adapts its source reconstructions to the properties of the image that it is
#    fitting, so as to objectively maximize the evidence!



from autolens.array import grids
from autolens.model.profiles import mass_profiles as mp
from autolens.model.profiles.plotters import profile_plotters

import numpy as np

grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(800, 800), pixel_scale=0.1
)

nfw_challenge = al.mass_profiles.SphericalTruncatedNFWChallenge(
    kappa_s=0.009, scale_radius=0.5
)
mass_at_200 = nfw_challenge.mass_at_200_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)
concentration = nfw_challenge.concentration_for_units(
    critical_surface_density=1940654909.4133248,
    cosmic_average_density=262.30319684750657,
)


print("{:.4e}".format(mass_at_200))
print(concentration)

deflections = nfw_challenge.deflections_from_grid(grid=grid)
print("defl y max = " + str(np.max(deflections[:, 0])))
print("defl x max = " + str(np.max(deflections[:, 1])))
print(
    "defl r max = "
    + str(np.max(np.sqrt(deflections[:, 1] ** 2.0 + deflections[:, 0] ** 2.0)))
)

profile_plotters.plot_deflections_x(mass_profile=nfw_challenge, grid=grid)
profile_plotters.plot_deflections_y(mass_profile=nfw_challenge, grid=grid)
